(*
GLUECUT (C) by Vladimir Kladov, 2003
VERSION 1.0, 24-Dec-2003
------------------------------------
This is a utility to convert KOL.PAS
created initially for Delphi2-7 for
Windows32 platform only to some other
versions compatible with other
compilers or/and target platforms.

KOL.PAS of version 1.89 or higher
required, containing section
separators and directives defined
below, and command file prepared.

Certainly, this utility can be used
with another source files too.

====================================
Usage:unpack GLUECUT.EXE to the same
directory where KOL.PAS is located,
and run it there (double-click in
the Explorer window, for example, if
a command file is named CMD or if
you want to enter name of CMD file
by a request).

Command line:

GLUECUT [/I:sourcefilepath]
        [/O:targetdirectory]
        [[/C[MD]:]commandfilename]

defaults:
/I:KOL.pas
/O:OUT
/C:CMD

default base directory to find input
and command files is a directory
from where GlueCut is started.

-------------------------------------
THEORY.

The source should have some labels
in form of //comment started in the
1st column of the line. Such labels
listed in the command file define
"sections" containing directives
certain for each section.
E.g.:

//: FILES DEFINITION :
...
//: REGISTRY DEFINITION :
...
//: TObj DEFINITION
...

A list of directives (may be yet not
complete), not case-sensitive:

SKIP [patternfrom] [patternto]
                - skip all the text
                between two patterns
                while copying

REPLACE [patternfrom] [textto]
                - replace all the
                occurances of
                [patternfrom] to the
                given [textto] (both
                without brackets)
                while copying

INCLUDE [filename]
                - include a file

TEXT --
  add a text until a line containing
  symbols "--" only (it is possible
  to use other pattern including one
  empty, in the last case an empty
  line is waiting for).
--

LINE            - inserts source file
                line number (and
                filename) into the
                output

COPY            - copy entire section
                to current output
                file

COPY [filename] - copy entire section
                to a filename. When
                the filename found
                the second (and
                further) time, the
                text is appended

#any comment

REWIND          - reset to the first
                line in the source
                and search following
                section separator

to skip entire section, just do not
use COPY in it. An order of commands
can be important, since COPY, INCLUDE
are executed immediately. All the lines
before the first section separator and
following the last one are copyied
"as is" using currently defined
output file. At the starting part,
a directive /OUT:directory\path can be
defined, which provides default output
directory (it can be overriden by the
option /O:directory in the command
line). In this part, a set of SKIP and
REPLACE directives can be defined which
has effect in all the sections below.
Such directives therefore are discarded
when REWIND executed, but another set
of such global commands can be defined
just following REWIND directive.

====================================
http: //bonanzas.rinet.ru
mailto: bonanzas@online.sinor.ru
*)

program GlueCut;

{$APPTYPE CONSOLE}

uses
  Windows,
  KOL;

function StrLEq( S1, S2: PChar; L: Integer ): Boolean;
asm
  PUSH ESI
  MOV  ESI, EAX
  CMP  ESI, EDX
  JZ   @@exit  // Z -> TRUE
  JECXZ @@exit // NZ -> FALSE
@@loop:
  LODSB
  MOV AH, AL
  OR  AH, [EDX]
  JZ  @@exit
  MOV AH, [EDX]
  INC EDX

  SUB AL, 'A'
  CMP AL, 'Z'-'A'
  JA  @@1
  ADD AL, $20
@@1:
  SUB AH, 'A'
  CMP AH, 'Z'-'A'
  JA  @@2
  ADD AH, $20
@@2:
  CMP AL, AH
  JNZ @@exit  // NZ -> FALSE
  LOOP @@loop
@@exit:
  SETZ AL
  POP  ESI
end;

{function StrLEq( S1, S2: PChar; L: Integer ): Boolean;
var A1, A2: String;
begin
  A1 := S1;
  A1 := Copy( A1, 1, L );
  A2 := S2;
  A2 := Copy( A2, 1, L );
  if (UpperCase( A1 ) = UpperCase( A2 )) and (A1 <> A2) then
  asm
    int 3;
  end;
  Result := StrLEq1( S1, S2, L );
end;}

procedure DoConversion;
const MaxBufLen = 1023;
var Line: Integer;
    BaseDir: String;
    OutDir: String;
    CurCmdIdx: Integer;

  procedure Error( const Msg: String );
  begin
    Writeln( 'Error in line ' + Int2Str( CurCmdIdx ) + ': ' + Msg
             //,MB_OK or MB_APPLMODAL or MB_SETFOREGROUND
           );
    Readln;
    Halt;
  end;

var ReplaceFrom, ReplaceTo: PStrList;
    SkipFrom, SkipTo: PStrList;
    Global_ReplaceFrom, Global_ReplaceTo: PStrList;
    Global_SkipFrom, Global_SkipTo: PStrList;
    OutputFiles: PStrList;
    F, InStream, OutStream, OutFile: PStream;
    Commands: PStrList;
    OK: Boolean;
    Buf: array[ 0..MaxBufLen ] of Char;
    BufPos: Integer;

    SkipWait: String;
    InputFileName: String;

  procedure SetOutput( const Filename: String );
  var I: Integer;
      S: String;
  begin
    I := OutputFiles.IndexOf( Filename );
    if (OutStream <> nil) and (OutStream.Position > 0) then
    begin
      if OutFile = nil then
      begin
        //        ,  :
        ForceDirectories( BaseDir + OutDir );
        S := BaseDir + OutDir + InputFileName;
        OutputFiles.Add( S );
        OutFile := NewWriteFileStream( S );
      end;
      OutStream.Size := OutStream.Position;
      OutStream.Position := 0;
      Stream2Stream( OutFile, OutStream, OutStream.Size );
    end;
    Free_And_Nil( OutFile );
    Free_And_Nil( OutStream );
    if Filename <> '' then
    begin
      if not ForceDirectories( ExtractFilePath( Filename ) ) then
      begin
        S := SysErrorMessage( GetLastError );
        AnsiToOem( PChar( S ), PChar( S ) );
        Error( 'Error creating directory: ' + ExtractFilePath( Filename ) + #13#10 +
               'Error code: ' + Int2Str( GetLastError ) + ' ' +
               S );
      end;
      if (I < 0) and (Filename <> '') then
      begin
        OutputFiles.Add( Filename );
        OutFile := NewWriteFileStream( Filename );
      end
        else
      begin
        OutFile := NewReadWriteFileStream( Filename );
        OutFile.Position := OutFile.Size;
      end;
      if OutFile.Handle = INVALID_HANDLE_VALUE then
      begin
        S := SysErrorMessage( GetLastError );
        AnsiToOem( PChar( S ), PChar( S ) );
        Error( 'Error writing output file: ' + Filename + #13#10 +
               'Error code: ' + Int2Str( GetLastError ) + ' ' +
               S );
      end;
    end;
    Writeln( 'Output to file: ' + Filename );
    OutStream := NewMemoryStream;
  end;

  procedure AddChar( Ch: Char );
  var I, L, L2: Integer;
  begin
    if OutStream = nil then
      SetOutput( OutDir + InputFilename );
    if OutStream.Size <= OutStream.Position + 1 then
      OutStream.Size := OutStream.Position + 100000;
    if (SkipWait = '') and (Ch <> #0) then
      OutStream.Write( Ch, 1 );
    if Ch < ' ' then
    begin
      BufPos := 0;
      Buf[ 0 ] := #0;
    end
      else
    begin
      if Ch <> #0 then
      begin
        Buf[ BufPos ] := Ch;
        Inc( BufPos );
      end;
      Buf[ BufPos ] := #0;
      if (BufPos >= MaxBufLen) or (Ch = #0) then
      begin
        Move( Buf[ 100 ], Buf[ 0 ], BufPos - 100 );
        Dec( BufPos, 100 );
      end;
      if SkipWait <> '' then
      begin
        L := Length( SkipWait );
        if L <= BufPos then
        if StrLEq( @ SkipWait[ 1 ], @ Buf[ BufPos - L ], L ) then
        begin
          BufPos := 0;
          Buf[ 0 ] := #0;
          SkipWait := '';
        end;
        Exit;
      end;
      if SkipWait = '' then
      begin
        for I := SkipFrom.Count-1 downto 0 do
        begin
          L := StrLen( SkipFrom.ItemPtrs[ I ] );
          if L <= BufPos then
          if StrLEq( SkipFrom.ItemPtrs[ I ],
                     @ Buf[ BufPos - L ], L ) then
          begin
            BufPos := 0;
            Buf[ 0 ] := #0;
            OutStream.Position := OutStream.Position - DWORD( L );
            SkipWait := SkipTo.Items[ I ];
            Exit;
          end;
        end;
      end;
      for I := ReplaceFrom.Count-1 downto 0 do
      begin
        L := StrLen( ReplaceFrom.ItemPtrs[ I ] );
        if (L <= BufPos) then
        if StrLEq( ReplaceFrom.ItemPtrs[ I ],
                   @ Buf[ BufPos - L ], L ) then
        begin
          L2 := StrLen( ReplaceTo.ItemPtrs[ I ] );
          Move( ReplaceTo.ItemPtrs[ I ]^, Buf[ 0 ], L2 + 1 );
          BufPos := L2;
          OutStream.Position := OutStream.Position - DWORD( L );
          OutStream.WriteStr( Buf );
          Exit;
        end;
      end;
    end;
  end;

  procedure AddStr( const S: String );
  var I: Integer;
  begin
    for I := 1 to Length( S ) do
      AddChar( S[ I ] );
    AddChar( #13 );
    AddChar( #10 );
  end;

  function CopySection( const EndStr: String; ToTheEnd: Boolean ): Boolean;
  var S: String;
  begin
    Result := FALSE;
    while TRUE do
    begin
      S := InStream.ReadStr;
      if not ToTheEnd and (TrimRight( S ) = TrimRight( EndStr )) then
      begin
        AddChar( #0 );
        SkipFrom.Assign( Global_SkipFrom );
        SkipTo.Assign( Global_SkipTo );
        ReplaceFrom.Assign( Global_ReplaceFrom );
        ReplaceTo.Assign( Global_ReplaceTo );
        AddStr( S );
        Result := TRUE;
        Exit;
      end
        else
      begin
        AddStr( S );
        if InStream.Position >= InStream.Size then break;
      end;
      if InStream.Position >= InStream.Size then Exit;
      Inc( Line );
    end;
  end;

  function SkipSection( const EndStr: String ): Boolean;
  var S: String;
  begin
    Result := FALSE;
    while TRUE do
    begin
      S := InStream.ReadStr;
      if TrimRight( S ) = TrimRight( EndStr ) then
      begin
        Result := TRUE;
        Exit;
      end;
      if InStream.Position >= InStream.Size then Exit;
      Inc( Line );
    end;
  end;

  procedure IncludeFile( Filename: String );
  var FileStrm: PStream;
      S: String;
  begin
    if not FileExists( Filename ) and FileExists( BaseDir + Filename ) then
      Filename := BaseDir + Filename;
    FileStrm := NewReadFileStream( Filename );
    if FileStrm.Handle = INVALID_HANDLE_VALUE then
      Error( 'File not found or could not open: ' + Filename );
    AddChar( #0 );
    while FileStrm.Position < FileStrm.Size do
    begin
      S := FileStrm.ReadStr;
      AddStr( S );
    end;
  end;

  function ParseParam( var S: String ): String;
  var W: String;
      I: Integer;
      R: Boolean;
  begin
    Parse( S, '[' );
    Result := '';
    while TRUE do
    begin
      W := Parse( S, ']' );
      R := TRUE;
      while R do
      begin
        R := FALSE;
        for I := 1 to Length( W )-1 do
          if (W[ I ] = '\') and (W[ I+1 ] = '\') then
          begin
            W := Copy( W, 1, I ) + CopyEnd( W, I+1 );
            R := TRUE;
            break;
          end;
      end;
      if (Length( W ) > 1) and (CopyTail( W, 2 ) = '\]') then
        Result := Result + Copy( W, 1, Length( W ) - 2 ) + ']'
      else
      begin
        Result := Result + W;
        break;
      end;
    end;

  end;

var S, Si: String;
    I, CmdIdx, NextSectionIdx: Integer;
    FirstSection: Boolean;
    OutDirDefined: Boolean;
    SectionTerminatedOK: Boolean;
begin
  Writeln( 'Converts source to a certain version using' +
           'given command file. Options:' );
  Writeln( '/I:inputfile - path is relative to current, by default /I:KOL.pas' );
  Writeln( '/O:outputdir - relative to current, by default /O:OUT' );
  Writeln( '/C[MD]:cmdfile - command file, by default /C:cmd' );

  S := '';
  OutDir := '';
  OutDirDefined := FALSE;
  for I := 1 to ParamCount do
  begin
    S := ParamStr( I );
    if StrIsStartingFrom( PChar( UpperCase( S ) ), '/I:' ) then
      S := CopyEnd( S, 4 )
    else
    if StrIsStartingFrom( PChar( UpperCase( S ) ), '/I' ) then
      S := CopyEnd( S, 3 )
    else
      S := '';
    if S <> '' then break;
  end;
  if S = '' then S := 'KOL.PAS';
  InputFileName := ExtractFileName( S );
  if pos( '\', S ) <= 0 then
  begin
    BaseDir := GetStartDir;
    S := BaseDir + S;
  end
    else
    BaseDir := ExtractFilePath( S );
  F := NewReadFileStream( S );
  if F.Handle = INVALID_HANDLE_VALUE then
    Error( 'Can not open input file: ' + S );
  InStream := NewMemoryStream;
  Stream2Stream( InStream, F, F.Size );
  InStream.Position := 0;
  F.Free;
  Global_ReplaceFrom := NewStrList;
  Global_ReplaceTo := NewStrList;
  Global_SkipFrom := NewStrList;
  Global_SkipTo := NewStrList;
  ReplaceFrom := NewStrList;
  ReplaceTo := NewStrList;
  SkipFrom := NewStrList;
  SkipTo := NewStrList;

  Commands := NewStrList;

  S := '';
  for I := 1 to ParamCount do
  begin
    S := ParamStr( I );
    if StrIsStartingFrom( PChar( UpperCase( S ) ), '/O:' ) then
      S := CopyEnd( S, 4 )
    else
    if StrIsStartingFrom( PChar( UpperCase( S ) ), '/O' ) then
      S := CopyEnd( S, 3 )
    else
      S := '';
    if S <> '' then break;
  end;
  if S = '' then
    S := 'OUT'
  else
    OutDirDefined := TRUE;
  OutDir := IncludeTrailingPathDelimiter( S );

  OutStream := nil;
  OutFile := nil;
  OutputFiles := NewStrList;

  S := '';
  for I := 1 to ParamCount do
  begin
    S := UpperCase( ParamStr( I ) );
    if StrIsStartingFrom( PChar( UpperCase( S ) ), '/CMD:' ) then
      S := CopyEnd( S, 6 )
    else
    if StrIsStartingFrom( PChar( UpperCase( S ) ), '/CMD' ) then
      S := CopyEnd( S, 5 )
    else
    if StrIsStartingFrom( PChar( UpperCase( S ) ), '/C:' ) then
      S := CopyEnd( S, 4 )
    else
    if StrIsStartingFrom( PChar( UpperCase( S ) ), '/C' ) then
      S := CopyEnd( S, 3 )
    else continue;
    break;
  end;
  if S = '' then
  begin
    for I := 1 to ParamCount do
    begin
      S := UpperCase( ParamStr( I ) );
      if Copy( S, 1, 1 ) <> '/' then
        break;
    end;
  end;
  if S = '' then
  begin
    while TRUE do
    begin
      Writeln( 'Enter commands file name: ' );
      Readln( S );
      if not FileExists( S ) and FileExists( BaseDir + S ) then
        S := BaseDir + S;
      if FileExists( S ) then break;
    end;
  end;

  if S <> '' then
  begin
    if not FileExists( S ) and FileExists( BaseDir + S ) then
      S := BaseDir + S
    else
    if not FileExists( S ) and FileExists( GetStartDir + S ) then
      S := GetStartDir + S;
    Commands.LoadFromFile( S );
    if Trim( Commands.Text ) = '' then
      Error( 'Command file is empty: ' + S );
  end;

  Writeln( 'Processing command file: ' + S );

  Line := 1;
  SkipWait := '';
  BufPos := 0;
  Buf[ 0 ] := #0;
  TRY
    //   
    CmdIdx := 0;
    while CmdIdx < Commands.Count do
    begin
      CurCmdIdx := CmdIdx;
      S := Commands.Items[ CmdIdx ];
      if StrIsStartingFrom( PChar( S ), '//' ) then break;
      if not OutDirDefined and ( UpperCase( Copy( S, 1, 5 ) ) = '/OUT:' ) then
      begin
        OutDirDefined := TRUE;
        OutDir := IncludeTrailingPathDelimiter( Trim( CopyEnd( S, 6 ) ) );
        if pos( ':', OutDir ) <= 0 then
          OutDir := GetStartDir + OutDir;
        SetOutput( OutDir + InputFileName );
      end
        else
      if StrIsStartingFrom( PChar( UpperCase( S ) ), 'SKIP' ) then
      begin
        S := Trim( CopyEnd( S, 5 ) );
        Global_SkipFrom.Add( ParseParam( S ) );
        Global_SkipTo.Add( ParseParam( S ) );
      end
        else
      if StrIsStartingFrom( PChar( UpperCase( S ) ), 'REPLACE' ) then
      begin
        S := Trim( CopyEnd( S, 8 ) );
        Global_ReplaceFrom.Add( ParseParam( S ) );
        Global_ReplaceTo.Add( ParseParam( S ) );
      end;
      Inc( CmdIdx );
    end;
    FirstSection := TRUE;
    while CmdIdx < Commands.Count do
    begin
      CurCmdIdx := CmdIdx;
      if FirstSection then //        
      begin
        FirstSection := FALSE;
        S := Commands.Items[ CmdIdx ];
        if not CopySection( S, FALSE ) then
          Error( 'Start of the first section not found: ' + S );
      end;

      //        
      NextSectionIdx := CmdIdx;
      while NextSectionIdx < Commands.Count do
      begin
        Inc( NextSectionIdx );
        S := Commands.Items[ NextSectionIdx ];
        if StrIsStartingFrom( PChar( S ), '//' ) then break;
        if StrIsStartingFrom( PChar( UpperCase( S ) ), 'TEXT' ) then
        begin
          S := Trim( CopyEnd( S, 5 ) );
          while TRUE do
          begin
            Inc( NextSectionIdx );
            if S = TrimRight( Commands.Items[ NextSectionIdx ] ) then
            begin
              //Inc( NextSectionIdx );
              break;
            end;
          end;
        end;
      end;

      //   ,  
      SectionTerminatedOK := FALSE;
      SkipFrom.Assign( Global_SkipFrom );
      SkipTo.Assign( Global_SkipTo );
      ReplaceFrom.Assign( Global_ReplaceFrom );
      ReplaceTo.Assign( Global_ReplaceTo );
      AddChar( #0 );
      I := CmdIdx;
      while I < NextSectionIdx-1 do
      begin
        Inc( I );
        CurCmdIdx := I;
        S := Commands.Items[ I ];
        if Copy( S, 1, 1 ) = '#' then //  ,  
        else
        if Trim( S ) = '' then //   , 
        else
        if StrIsStartingFrom( PChar( UpperCase( S ) ), 'COPY' ) then
        begin
          S := Trim( CopyEnd( S, 5 ) );
          if S <> '' then
            SetOutput( S );
          //   
          S := '';
          if NextSectionIdx < Commands.Count then
            S := TrimRight( Commands.Items[ NextSectionIdx ] );
          SectionTerminatedOK :=
            CopySection( S, NextSectionIdx >= Commands.Count );
          if not SectionTerminatedOK then
            Error( 'Section separator not found: ' + S );
        end
          else
        if StrIsStartingFrom( PChar( UpperCase( S ) ), 'SKIP' ) then
        begin
          S := Trim( CopyEnd( S, 5 ) );
          SkipFrom.Add( ParseParam( S ) );
          SkipTo.Add( ParseParam( S ) );
        end
          else
        if StrIsStartingFrom( PChar( UpperCase( S ) ), 'REPLACE' ) then
        begin
          S := Trim( CopyEnd( S, 8 ) );
          ReplaceFrom.Add( ParseParam( S ) );
          ReplaceTo.Add( ParseParam( S ) );
        end
          else
        if StrIsStartingFrom( PChar( UpperCase( S ) ), 'INCLUDE' ) then
        begin
          S := Trim( CopyEnd( S, 8 ) );
          S := ParseParam( S );
          AddChar( #0 );
          IncludeFile( S );
        end
          else
        if StrIsStartingFrom( PChar( UpperCase( S ) ), 'TEXT' ) then
        begin
          S := Trim( CopyEnd( S, 5 ) );
          while TRUE do
          begin
            Inc( I );
            Si := Commands.Items[ I ];
            if TrimRight( Si ) = S then
            begin
              //Inc( I );
              break;
            end;
            AddStr( Si );
          end;
        end
          else
        if StrIsStartingFrom( PChar( UpperCase( S ) ), 'LINE' ) then
        begin
          S := '// ' + InputFilename + ' Line ' + Int2Str( Line );
          AddChar( #0 );
          OutStream.WriteStr( S + #13#10 );
        end
          else
        if StrEq( S, 'REWIND' ) then
        begin //   
          InStream.Position := 0;
          S := '';
          if NextSectionIdx < Commands.Count then
            S := TrimRight( Commands.Items[ NextSectionIdx ] );
          SectionTerminatedOK :=
            SkipSection( S );
          if not SectionTerminatedOK then
            Error( 'Section separator not found: ' + S );
          //   SKIP [from][to]  REPLACE [from][to]  
          // REWIND  
          while I < NextSectionIdx-1 do
          begin
            Inc( I );
            CurCmdIdx := I;
            S := Commands.Items[ I ];
            if Copy( S, 1, 1 ) = '#' then //  ,  
            else
            if Trim( S ) = '' then //   , 
            else
            if StrIsStartingFrom( PChar( UpperCase( S ) ), 'SKIP' ) then
            begin
              S := Trim( CopyEnd( S, 5 ) );
              Global_SkipFrom.Add( ParseParam( S ) );
              Global_SkipTo.Add( ParseParam( S ) );
            end
              else
            if StrIsStartingFrom( PChar( UpperCase( S ) ), 'REPLACE' ) then
            begin
              S := Trim( CopyEnd( S, 8 ) );
              Global_ReplaceFrom.Add( ParseParam( S ) );
              Global_ReplaceTo.Add( ParseParam( S ) );
            end
              else
              Error( 'Unknown command (after REWIND only SKIP and ' +
                     'REPLACE are allowed):' + S );
          end;
        end
          else
          Error( 'Unknown command: ' + S );
      end;
      if not SectionTerminatedOK and (NextSectionIdx < Commands.Count) then
      begin
        S := TrimRight( Commands.Items[ NextSectionIdx ] );
        SectionTerminatedOK := SkipSection( S );
        if not SectionTerminatedOK then
          Error( 'Section separator not found: ' + S );
      end;
      CmdIdx := NextSectionIdx;
    end;
    CopySection( '', TRUE );
    OK := TRUE;
  FINALLY
    ReplaceFrom.Free;
    ReplaceTo.Free;
    SkipFrom.Free;
    SkipTo.Free;
    Commands.Free;
    OutputFiles.Free;
    AddChar( #0 );
    SetOutput( '' );
  END;
  if not OK then
    Error( 'Can not write file ' + S + ': ' + SysErrorMessage( GetLastError ) )
  else
    Writeln( InputFileName + ' converted OK.' );
end;

begin
  UseDelphiMemoryManager;
  UseInputOutput;
  DoConversion;
end.
